#pragma once

// 包含必要的公共头文件和依赖模块
#include "common/common.h"
#include "oh_my_loam/base/feature.h"
#include "oh_my_loam/base/utils.h"
#include "oh_my_loam/base/types.h"
#include "oh_my_loam/visualizer/extractor_visualizer.h"

namespace oh_my_loam {

/**
 * @brief 特征提取器基类
 *
 * 本类定义了点云特征提取的公共接口和基本实现，包括：
 *  - 初始化（Init）
 *  - 处理输入点云数据并提取特征（Process）
 *  - 分割扫描线、计算曲率、分配点类型以及生成特征点等步骤
 *
 * 注意：具体的扫描线编号获取方法（GetScanID）由派生类根据激光雷达的具体结构实现。
 */
class Extractor {
 public:
  // 默认构造函数
  Extractor() = default;
  // 默认虚析构函数，确保派生类析构时正确释放资源
  virtual ~Extractor() = default;

  /**
   * @brief 初始化特征提取器
   *
   * 读取配置文件中与特征提取相关的参数，并进行必要的初始化设置，
   * 如是否开启可视化等。
   *
   * @return true 初始化成功
   */
  bool Init();

  /**
   * @brief 处理输入点云数据，提取特征
   *
   * 根据输入的点云数据，依次执行以下步骤：
   *  - 将点云分割为多个扫描线
   *  - 计算每个扫描线中各点的曲率
   *  - 根据曲率为点分配类型（例如角点或平面点）
   *  - 根据不同的点类型生成特征点集合
   *  - （可选）进行可视化展示
   *
   * @param timestamp 当前帧的时间戳
   * @param cloud 输入的点云数据（常量指针）
   * @param features 输出的特征集合，包含角点、平面点等
   */
  void Process(double timestamp, const common::PointCloudConstPtr &cloud,
               std::vector<Feature> *const features);

  /**
   * @brief 获取激光扫描线的数量
   *
   * 返回激光雷达的扫描线数，该值通常由配置文件指定。
   *
   * @return int 扫描线数量
   */
  int num_scans() const {
    return num_scans_;
  }

  /**
   * @brief 重置提取器状态
   *
   * 虚函数，允许派生类根据需要重置内部状态（例如清空缓存）。
   */
  virtual void Reset() {}

 protected:
  /**
   * @brief 获取点所属的扫描线ID
   *
   * 根据输入点的坐标信息，计算该点属于哪一条激光扫描线。
   * 此函数为纯虚函数，必须由具体的雷达类型派生类实现。
   *
   * @param pt 输入点
   * @return int 扫描线ID
   */
  virtual int GetScanID(const common::Point &pt) const = 0;

  /**
   * @brief 分割点云为多个扫描线
   *
   * 根据输入的点云数据和激光雷达扫描线数量，将点云数据按照水平角度等信息
   * 分割到各个扫描线中，为后续计算曲率和提取特征做准备。
   *
   * @param cloud 输入的点云数据
   * @param scans 输出的扫描线集合，每个扫描线存储在一个 TCTPointCloud 对象中
   */
  virtual void SplitScan(const common::PointCloud &cloud,
                         std::vector<TCTPointCloud> *const scans) const;

  /**
   * @brief 计算扫描线中每个点的曲率
   *
   * 针对每条扫描线，利用点的邻域信息计算局部曲率，曲率反映了点的局部
   * 平面性或角点特性，是后续分配点类型的重要依据。
   *
   * @param scan 输入的扫描线数据，计算后会更新每个点的曲率值
   */
  virtual void ComputeCurvature(TCTPointCloud *const scan) const;

  /**
   * @brief 根据曲率为扫描线中的点分配类型
   *
   * 根据计算的曲率以及预设阈值，将扫描线中的点分配为不同类型，
   * 例如角点（包括尖锐角点与一般角点）和表面点（平面点）。
   *
   * @param scan 输入的扫描线数据，处理后每个点的类型会被更新
   */
  virtual void AssignType(TCTPointCloud *const scan) const;

  /**
   * @brief 根据扫描线生成特征点集合
   *
   * 遍历已分配好类型的扫描线，将点根据其类型分别存入不同的特征集合中，
   * 如角点和平面点，并可对平面点进行降采样处理。
   *
   * @param scan 输入的扫描线数据（已分配类型）
   * @param feature 输出的特征数据结构，包含各类特征点
   */
  virtual void GenerateFeature(const TCTPointCloud &scan,
                               Feature *const feature) const;

  /**
   * @brief 可视化提取的特征
   *
   * 将当前帧的原始点云和提取的特征数据封装成一个视觉化帧，
   * 通过视觉化工具进行渲染，便于调试和结果展示。
   *
   * @param cloud 当前帧的原始点云数据
   * @param features 当前帧提取的特征集合
   * @param timestamp 当前帧的时间戳（默认为 0.0）
   */
  virtual void Visualize(const common::PointCloudConstPtr &cloud,
                         const std::vector<Feature> &features,
                         double timestamp = 0.0) const;

  // 激光雷达扫描线的数量，通常由配置文件中设置
  int num_scans_ = 0;

  // 存储特征提取相关的配置参数（YAML 格式）
  YAML::Node config_;

  // 指向特征提取结果可视化工具的智能指针
  std::unique_ptr<ExtractorVisualizer> visualizer_{nullptr};

  // 是否打印详细日志信息的标志
  bool verbose_ = false;

 private:
  /**
   * @brief 更新邻域点的标记状态
   *
   * 当某个点被选为特征点后，为避免其邻域内的点过于密集地被选取，
   * 将该点附近一定范围内的点标记为已选。
   *
   * @param scan 当前扫描线数据
   * @param ix 当前选中点在扫描线中的索引
   * @param picked 标记数组，记录每个点是否已经被选取
   */
  void UpdateNeighborsPicked(const TCTPointCloud &scan, int ix,
                             std::vector<bool> *const picked) const;

  // 标志是否开启可视化，true 表示开启
  bool is_vis_ = false;

  // 禁止拷贝和赋值操作的宏定义，确保对象不可拷贝
  DISALLOW_COPY_AND_ASSIGN(Extractor);
};

}  // namespace oh_my_loam
