#pragma once

// 包含必要的公共头文件和模块
#include "common/common.h"
#include "oh_my_loam/base/feature.h"
#include "oh_my_loam/base/utils.h"
#include "oh_my_loam/solver/solver.h"
#include "oh_my_loam/visualizer/odometer_visualizer.h"

namespace oh_my_loam {

/**
 * @brief 基于帧间匹配的激光雷达里程计
 *
 * 该类实现了激光雷达帧间位姿估计，主要功能包括：
 *  - 初始化参数和可视化组件（Init）
 *  - 根据当前帧提取的特征计算相对于上一帧以及全局的位姿更新（Process）
 *  - 角点与平面点的匹配、最近邻搜索及结果可视化等辅助功能
 */
class Odometer {
 public:
  // 默认构造函数
  Odometer() = default;

  /**
   * @brief 初始化里程计
   *
   * 读取配置文件中的参数，初始化各项内部变量，并根据配置决定是否启用可视化。
   *
   * @return true 初始化成功
   */
  bool Init();

  /**
   * @brief 处理当前帧特征，更新里程计位姿
   *
   * 根据输入的特征集合（如角点与平面点），通过匹配和求解算法估计当前帧相对于上一帧的位姿，
   * 并更新全局位姿。
   *
   * @param timestamp 当前帧时间戳
   * @param features 当前帧的特征集合
   * @param pose_out 输出的全局位姿
   */
  void Process(double timestamp, const std::vector<Feature> &features,
               common::Pose3d *const pose_out);

  /**
   * @brief 获取当前角点点云
   *
   * 返回存储在全局角点 kd-tree 中的角点点云数据。
   *
   * @return TPointCloudConstPtr 角点点云的常量指针
   */
  TPointCloudConstPtr GetCloudCorn() const {
    return kdtree_corn_.getInputCloud();
  }

  /**
   * @brief 获取当前平面点点云
   *
   * 返回存储在全局平面点 kd-tree 中的平面点点云数据。
   *
   * @return TPointCloudConstPtr 平面点点云的常量指针
   */
  TPointCloudConstPtr GetCloudSurf() const {
    return kdtree_surf_.getInputCloud();
  }

  /**
   * @brief 重置里程计状态
   *
   * 重置内部状态变量，用于系统重启或丢失跟踪后的恢复。
   */
  void Reset();

 protected:
  /**
   * @brief 更新上一帧特征数据
   *
   * 将当前帧的特征数据转换后存储，作为下一帧匹配的参考数据。
   *
   * @param features 当前帧的特征集合
   */
  void UpdatePre(const std::vector<Feature> &features);

  /**
   * @brief 匹配角点，构造点-直线匹配对
   *
   * 在输入的角点点云中搜索匹配点，并构造点-直线匹配对，
   * 该匹配对将用于后续位姿求解。
   *
   * @param src 输入的角点点云
   * @param pairs 输出的匹配结果（点-直线匹配对集合）
   */
  void MatchCorn(const TPointCloud &src,
                 std::vector<PointLinePair> *const pairs) const;

  /**
   * @brief 匹配平面点，构造点-平面对
   *
   * 在输入的平面点点云中搜索匹配点，并构造点-平面对，
   * 该匹配对将用于后续位姿求解。
   *
   * @param src 输入的平面点点云
   * @param pairs 输出的匹配结果（点-平面对集合）
   */
  void MatchSurf(const TPointCloud &src,
                 std::vector<PointPlanePair> *const pairs) const;

  /**
   * @brief 可视化匹配结果与位姿信息
   *
   * 利用可视化工具将当前匹配的点对以及位姿信息进行展示，
   * 便于调试和监控系统运行状态。
   *
   * @param pl_pairs 点-直线匹配对集合
   * @param pp_pairs 点-平面对集合
   * @param timestamp 当前时间戳（默认为 0.0）
   */
  void Visualize(const std::vector<PointLinePair> &pl_pairs,
                 const std::vector<PointPlanePair> &pp_pairs,
                 double timestamp = 0.0) const;

  /**
   * @brief 辅助的最近邻搜索函数
   *
   * 在给定的 kd-tree 中查找查询点的 k 个最近邻，如果距离平方满足阈值条件，
   * 则返回对应的索引集合。
   *
   * @param kdtree 用于搜索的 kd-tree 对象
   * @param query_pt 查询点
   * @param k 搜索的最近邻个数
   * @param dist_sq_th 距离平方阈值
   * @param indices 输出的最近邻索引集合
   * @return true 搜索成功，false 否则
   */
  bool NearestSearch(const pcl::KdTreeFLANN<TPoint> &kdtree,
                     const TPoint &query_pt, int k, float dist_sq_th,
                     std::vector<int> *const indices) const;

  // 当前帧到世界坐标系的全局位姿
  common::Pose3d pose_curr2world_;
  // 当前帧相对于上一帧的位姿变换
  common::Pose3d pose_curr2last_;

  // 存储每个扫描线的平面点 kd-tree，用于加速平面点匹配
  std::vector<pcl::KdTreeFLANN<TPoint>> kdtrees_scan_surf_;
  // 存储每个扫描线的角点 kd-tree，用于加速角点匹配
  std::vector<pcl::KdTreeFLANN<TPoint>> kdtrees_scan_corn_;
  // 全局角点 kd-tree（用于全局角点匹配）
  pcl::KdTreeFLANN<TPoint> kdtree_corn_;
  // 全局平面点 kd-tree（用于全局平面点匹配）
  pcl::KdTreeFLANN<TPoint> kdtree_surf_;

  // 存储里程计相关的配置参数（例如匹配距离阈值、迭代次数等）
  YAML::Node config_;

  // 里程计是否已初始化
  bool is_initialized_ = false;

  // 是否启用可视化
  bool is_vis_ = false;

  // 是否输出详细日志信息
  bool verbose_ = false;

  // 里程计可视化工具（用于展示匹配和位姿估计结果）
  std::unique_ptr<OdometerVisualizer> visualizer_{nullptr};

  // 禁止拷贝构造函数和赋值操作，防止对象被误拷贝
  DISALLOW_COPY_AND_ASSIGN(Odometer)
};

}  // namespace oh_my_loam
