#include "solver.h"

#include <Eigen/src/Core/Matrix.h>
#include <ceres/loss_function.h>
#include <ceres/types.h>

#include "common/log/log.h"

namespace oh_my_loam {

// 定义一个局部变量，表示 Huber 损失的尺度参数
namespace {
double kHuberLossScale = 0.1;
}

/**
 * @brief PoseSolver 构造函数
 *
 * 初始化位姿求解器，传入初始位姿（包括旋转和位移），并将其复制到内部参数数组中。
 * 同时为 Ceres 问题添加参数块，其中旋转部分使用 Eigen 四元数参数化，
 * 并设置 Huber 损失函数用于鲁棒优化。
 *
 * @param pose 初始位姿（类型为 common::Pose3d），包含旋转（四元数）和位移向量
 */
PoseSolver::PoseSolver(const common::Pose3d &pose) {
  // 将输入位姿的旋转部分（四元数的系数）复制到内部数组 r_quat_ 中，四元数有 4 个元素
  std::copy_n(pose.r_quat().coeffs().data(), 4, r_quat_);
  // 将输入位姿的平移向量复制到内部数组 t_vec_ 中，平移向量有 3 个元素
  std::copy_n(pose.t_vec().data(), 3, t_vec_);
  // 创建一个 Ceres 的 Huber 损失函数对象，参数 kHuberLossScale 控制损失函数的尺度
  loss_function_ = new ceres::HuberLoss(kHuberLossScale);
  // 向问题中添加旋转参数块，使用 Eigen 四元数参数化以确保单位四元数的约束
  problem_.AddParameterBlock(r_quat_, 4, new ceres::EigenQuaternionParameterization());
  // 向问题中添加平移参数块（3个自由度）
  problem_.AddParameterBlock(t_vec_, 3);
}

/**
 * @brief Solve 求解器入口函数
 *
 * 利用 Ceres 求解器求解当前的位姿优化问题。设定线性求解器类型、最大迭代次数等参数，
 * 求解后更新内部的旋转和平移参数，同时将最新位姿保存到 pose 输出参数中。
 *
 * @param max_iter_num 最大迭代次数
 * @param verbose 是否输出详细日志信息
 * @param pose 求解后得到的最新位姿（输出参数）
 * @return bool 如果优化收敛则返回 true，否则返回 false
 */
bool PoseSolver::Solve(int max_iter_num, bool verbose,
                       common::Pose3d *const pose) {
  // 配置 Ceres 求解选项
  ceres::Solver::Options options;
  options.linear_solver_type = ceres::DENSE_QR;  // 使用密集 QR 分解作为线性求解器
  options.max_num_iterations = max_iter_num;       // 最大迭代次数
  options.minimizer_progress_to_stdout = false;    // 是否将求解进度输出到标准输出
  ceres::Solver::Summary summary;                  // 用于存储求解过程和结果的摘要信息
  // 调用 Ceres 求解器求解问题
  ceres::Solve(options, &problem_, &summary);
  // 如果 verbose 为 true，则输出求解摘要报告
  AINFO_IF(verbose) << summary.BriefReport();
  // 如果 pose 非空，则将当前求解得到的旋转和平移参数构造为 common::Pose3d 对象输出
  if (pose) *pose = common::Pose3d(r_quat_, t_vec_);
  // 返回求解是否收敛的状态（Ceres 的 termination_type 为 CONVERGENCE 表示收敛）
  return summary.termination_type == ceres::CONVERGENCE;
}

/**
 * @brief 添加点-直线匹配项
 *
 * 根据输入的点-直线匹配对 pair 以及时间参数 time，
 * 构造一个对应的代价函数（CostFunction），并将其加入到 Ceres 优化问题中。
 *
 * @param pair 点-直线匹配对，包含原始点和匹配直线上的两个点
 * @param time 时间参数，用于对匹配代价进行时间权重调整
 */
void PoseSolver::AddPointLinePair(const PointLinePair &pair, double time) {
  // 通过静态工厂函数创建点-直线匹配的代价函数
  ceres::CostFunction *cost_function =
      PointLineCostFunction::Create(pair, time);
  // 将代价函数添加到问题中，设置损失函数（Huber 损失）和待优化参数（旋转和平移）
  problem_.AddResidualBlock(cost_function, loss_function_, r_quat_, t_vec_);
}

/**
 * @brief 添加点-平面对匹配项
 *
 * 根据输入的点-平面对匹配 pair 以及时间参数 time，
 * 构造一个对应的代价函数（CostFunction），并将其加入到 Ceres 优化问题中。
 *
 * @param pair 点-平面对匹配对，包含原始点和匹配平面上的三个点
 * @param time 时间参数，用于对匹配代价进行时间权重调整
 */
void PoseSolver::AddPointPlanePair(const PointPlanePair &pair, double time) {
  // 创建点-平面对代价函数
  ceres::CostFunction *cost_function =
      PointPlaneCostFunction::Create(pair, time);
  // 添加残差块到优化问题中，待优化参数为旋转和平移
  problem_.AddResidualBlock(cost_function, loss_function_, r_quat_, t_vec_);
}

/**
 * @brief 添加带有系数信息的点-直线匹配项
 *
 * 针对带有额外系数信息的点-直线匹配对，构造对应的代价函数，并加入优化问题中。
 *
 * @param pair 点-直线匹配对（含系数信息）
 * @param time 时间参数
 */
void PoseSolver::AddPointLineCoeffPair(const PointLineCoeffPair &pair,
                                       double time) {
  // 通过工厂函数创建带系数的点-直线代价函数
  ceres::CostFunction *cost_function =
      PointLineCoeffCostFunction::Create(pair, time);
  // 将代价函数加入到问题中
  problem_.AddResidualBlock(cost_function, loss_function_, r_quat_, t_vec_);
}

/**
 * @brief 添加带有系数信息的点-平面对匹配项
 *
 * 针对带有额外系数信息的点-平面对匹配对，构造对应的代价函数，并加入优化问题中。
 *
 * @param pair 点-平面对匹配对（含系数信息）
 * @param time 时间参数
 */
void PoseSolver::AddPointPlaneCoeffPair(const PointPlaneCoeffPair &pair,
                                        double time) {
  // 创建带系数的点-平面代价函数
  ceres::CostFunction *cost_function =
      PointPlaneCoeffCostFunction::Create(pair, time);
  // 将代价函数添加到 Ceres 问题中，优化参数为旋转和平移
  problem_.AddResidualBlock(cost_function, loss_function_, r_quat_, t_vec_);
}

/**
 * @brief 获取当前求解的位姿
 *
 * 根据内部存储的旋转和平移参数构造一个 common::Pose3d 对象，
 * 用于返回当前的位姿估计结果。
 *
 * @return common::Pose3d 当前位姿
 */
common::Pose3d PoseSolver::GetPose() const {
  return common::Pose3d(r_quat_, t_vec_);
}

}  // namespace oh_my_loam
